package logwire.web.bo.loader;

import logwire.core.bo.annotation.*;
import logwire.core.bo.field.BizArray;
import logwire.core.bo.field.BizMany;
import logwire.core.bo.field.BizText;
import logwire.core.bo.object.BizObject;
import logwire.core.exceptions.ApplicationException;
import logwire.core.meta.model.Model;
import logwire.core.meta.model.fields.*;
import logwire.core.meta.query.IQuery;
import logwire.core.resource.BeanLoader;
import logwire.core.resource.loader.JavaModelBeanLoader;
import logwire.core.tenant.TenantClassLoader;
import logwire.core.tenant.TenantProject;
import logwire.web.bo.field.parse.FieldParser;
import logwire.web.bo.field.parse.FieldParserUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.core.io.DefaultResourceLoader;

import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;
import java.util.stream.Collectors;

public class BoModelBeanLoader extends BeanLoader<IQuery, TenantProject> implements JavaModelBeanLoader {

    @Autowired
    LoaderUtil loaderUtil;

    /**
     * load BoModel 的逻辑：
     * 1、先遍历出所有 BizObject 的子类 得到集合
     * 2、筛选出当中使用 @BizModel 的类，构建 Model，字段包含父类的字段
     * 3、将集合中的剩余类的字段添加到父类对应的Model 中
     *
     * @param project  当前租户
     * @param input    待加载的资源
     * @param consumer 处理加载结果
     * @throws Exception
     */
    @Override
    public void load(TenantProject project, TenantProject input, Consumer<IQuery> consumer) throws Exception {
        TenantClassLoader classLoader = project.getTenantBeanContext().getTenantClassLoader();

        //先扫描到所有的 BizObject (BizObject的所有子类)
        ClassPathScanningCandidateComponentProvider provider = loaderUtil.getScanningBeanProvider();
        provider.setResourceLoader(new DefaultResourceLoader(classLoader));
        provider.addIncludeFilter(loaderUtil.getAbstractTypeFilter(BizObject.class));
        Map<Boolean, List<BeanDefinition>> isBizModel = provider.findCandidateComponents("logwire.web")
                .stream()
                .collect(Collectors.groupingBy(beanDefinition -> {
                    try {
                        Class<?> clazz = classLoader.loadClass(beanDefinition.getBeanClassName());
                        BizModel bizModelAnnotation = clazz.getAnnotation(BizModel.class);
                        return bizModelAnnotation != null;
                    } catch (ClassNotFoundException e) {
                        throw new ApplicationException("");
                    }
                }));
        List<BeanDefinition> bizModelList = isBizModel.get(true);
        for (BeanDefinition beanDefinition : bizModelList) {
            Class<?> clazz = classLoader.loadClass(beanDefinition.getBeanClassName());
            //独立model
            Model model = new Model();
            model.setName(clazz.getSimpleName());
            model.setVerboseName(clazz.getAnnotation(BizModel.class).label());
            model.setIncludeAuditFields(true);
            model.setIncludeVersionField(true);
            model.setIncludeDomainField(true);
            initField(clazz, model, consumer);
            consumer.accept(model);

        }


    }

    private void initField(Class<?> clazz, Model model, Consumer consumer) {
        Field[] fields = clazz.getDeclaredFields();
        for (Field f : fields) {
            if (f.getAnnotation(Column.class) != null) {
                //带有Column注解的字段
                FieldParser parser = FieldParserUtil.getParser(f);
                logwire.core.meta.model.fields.Field field = parser.parse(f.getName(), f);
                model.addField(field);
            } else if (f.getAnnotation(Text.class) != null) {
                // 带有Text注解的字段 大文本属性会创建一张子表(Model),子表名为：BO表名_属性名，一个BO有多个大字段时，使用同一张表
                if (f.getType() != BizText.class) {
                    throw new ApplicationException("@Text 字段类型必须为BizText");
                }
                newTextModel(model, f, consumer);
            } else if (f.getAnnotation(Array.class) != null) {
                // 带有Text注解的字段 大文本属性会创建一张子表(Model),子表名为：BO表名_属性名，一个BO有多个大字段时，使用同一张表
                if (f.getType() != BizArray.class) {
                    throw new ApplicationException("@Array 字段类型必须为BizArray");
                }
                newArrayModel(model, f, consumer);
            } else if (f.getAnnotation(Many.class) != null) {
                // 带有Text注解的字段 大文本属性会创建一张子表(Model),子表名为：BO表名_属性名，一个BO有多个大字段时，使用同一张表
                if (f.getType() != BizMany.class) {
                    throw new ApplicationException("@Many 字段类型必须为BizMany");
                }
                newManyModel(model, f, consumer);
            }
        }
    }

    /**
     * 数组字段值存到一个列中，用逗号分隔 字段：
     * owner_id/所属对象ID   (BigInt)
     * value/值     (varchar/numeric/boolean/bigint/integer/timestamp)
     * idx/顺序 Integer
     *
     * @param f
     * @param consumer
     */
    private void newArrayModel(Model model, Field f, Consumer consumer) {
        Array array = f.getAnnotation(Array.class);
        Model arrayModel = new Model();
        arrayModel.setName(model.getName() + "_" + f.getName());
        arrayModel.setVerboseName(array.label());
        ForeignKeyField ownerId = new ForeignKeyField("owner_id", model.getName());
        ownerId.setPrimaryKey(true);
        arrayModel.addField(ownerId);
        arrayModel.addField(new IntegerField("idx"));
        Type actualTypeArguments = ((ParameterizedType) f.getGenericType()).getActualTypeArguments()[0];
        FieldParser parser = FieldParserUtil.getParserByFieldType(actualTypeArguments);
        logwire.core.meta.model.fields.Field valueField = parser.parse("value", f);
        model.addField(valueField);
        consumer.accept(model);
    }

    /**
     * 大文本 字段：
     * owner_id/所属对象ID   (BigInt)
     * field_name/字段名 varchar(50)
     * value/值     (varchar/numeric/boolean/bigint/integer/timestamp)
     *
     * @param model
     * @param f
     * @param consumer
     */
    private void newTextModel(Model model, Field f, Consumer consumer) {
        Text text = f.getAnnotation(Text.class);
        Model textModel = new Model();
        textModel.setName(model.getName() + "_text");
        textModel.setVerboseName(text.label());
        ForeignKeyField ownerId = new ForeignKeyField("owner_id", model.getName());
        ownerId.setPrimaryKey(true);
        textModel.addField(ownerId);
        textModel.addField(new StringField("field_name", 50));
        textModel.addField(new TextField("value"));
        consumer.accept(model);
    }

    /**
     * 多对多关联属性 字段：
     * id
     * master_id 主表ID
     * slave_id  从表ID
     *
     * @param model
     * @param f
     * @param consumer
     */
    private void newManyModel(Model model, Field f, Consumer consumer) {
        Many many = f.getAnnotation(Many.class);
        Model manyModel = new Model();
        manyModel.setName(model.getName() + "_" + f.getName());
        manyModel.setVerboseName(many.label());
        manyModel.addField(new AutoField());
        manyModel.addField(new ForeignKeyField("master_id", model.getName()));
        manyModel.addField(new ForeignKeyField("slave_id", many.referBoName()));
        consumer.accept(model);
    }

    @Override
    public boolean accept(TenantProject input) {
        return true;
    }
}
